Package: dcortools
Title: Providing Fast and Flexible Functions for Distance Correlation
        Analysis
Description: Provides methods for distance covariance and distance correlation (Szekely, et al. (2007) <doi:10.1214/009053607000000505>), generalized version thereof (Sejdinovic, et al. (2013) <doi:10.1214/13-AOS1140>) and corresponding tests (Berschneider, Bottcher (2018) <doi:10.48550/arXiv.1808.07280>. Distance standard deviation methods (Edelmann, et al. (2020) <doi:10.1214/19-AOS1935>) and distance correlation methods for survival endpoints (Edelmann, et al. (2021) <doi:10.1111/biom.13470>) are also included.
Version: 0.2.1
Authors@R: c(person("Dominic", "Edelmann", email = "dominic.edelmann@dkfz-heidelberg.de", role = c("aut", "cre")), person("Jochen", "Fiedler", email = "jochen.fiedler@itwm.fraunhofer.de", role = c("aut")))
Depends: Rdpack, R (>= 3.3.3), Rcpp (>= 0.11.0), Rfast
Imports: ggplot2, pheatmap, Hmisc, stats, RSpectra, CompQuadForm,
        survey
RdMacros: Rdpack
LinkingTo: RcppArmadillo, Rcpp, RcppEigen
License: GPL-3
Encoding: UTF-8
NeedsCompilation: yes
Author: Dominic Edelmann [aut, cre],
  Jochen Fiedler [aut]
Maintainer: Dominic Edelmann <dominic.edelmann@dkfz-heidelberg.de>
Repository: CRAN
Config/roxygen2/version: 8.0.0
Packaged: 2026-07-16 18:52:48 UTC; edelmand
Date/Publication: 2026-07-16 19:30:02 UTC
