# Generated by roxygen2: do not edit by hand

S3method(plot,dcmatrix)
S3method(print,dcmatrix)
S3method(print,dctest)
export(dcmatrix)
export(dcorgaussianbiv)
export(dcsis)
export(distcor)
export(distcov)
export(distcov.test)
export(distsd)
export(hsplot)
export(ipcw.dcor)
export(ipcw.dcov)
export(ipcw.dcov.test)
importFrom(Rcpp,sourceCpp)
importFrom(Rdpack,reprompt)
importFrom(Rfast,standardise)
useDynLib(dcortools, .registration = TRUE)
