dcortools R package 
####################
by Dominic Edelmann (dominic.edelmann@dkfz-heidelberg.de) and Jochen Fiedler

This package provides functionalities for calculating association measures and 
dependence tests based on distance covariance and distance correlation.

