# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

squaresym <- function(A) {
    .Call(`_dcortools_squaresym`, A)
}

normalize_matrix <- function(M, cmM, mM) {
    .Call(`_dcortools_normalize_matrix`, M, cmM, mM)
}

hadamard_product <- function(X, Y) {
    .Call(`_dcortools_hadamard_product`, X, Y)
}

matrix_prod_sum2 <- function(X, Y) {
    .Call(`_dcortools_matrix_prod_sum2`, X, Y)
}

sum_hadamard_power2 <- function(X) {
    .Call(`_dcortools_sum_hadamard_power2`, X)
}

sum_hadamard_power3 <- function(X) {
    .Call(`_dcortools_sum_hadamard_power3`, X)
}

vector_product <- function(X, Y) {
    .Call(`_dcortools_vector_product`, X, Y)
}

matrix_prod_sum <- function(X, Y) {
    .Call(`_dcortools_matrix_prod_sum`, X, Y)
}

vector_prod_sum <- function(X, Y) {
    .Call(`_dcortools_vector_prod_sum`, X, Y)
}

specific_vector_prod_sum <- function(X, Y, gamma_1, gamma_X, gamma_Y, gamma_XY) {
    .Call(`_dcortools_specific_vector_prod_sum`, X, Y, gamma_1, gamma_X, gamma_Y, gamma_XY)
}

matrix_sum <- function(X) {
    .Call(`_dcortools_matrix_sum`, X)
}

matrix_prod_sum_sample <- function(X, Y, s) {
    .Call(`_dcortools_matrix_prod_sum_sample`, X, Y, s)
}

matrix_prod_sum_subset <- function(X, Y, s) {
    .Call(`_dcortools_matrix_prod_sum_subset`, X, Y, s)
}

colsums_subset <- function(X, s) {
    .Call(`_dcortools_colsums_subset`, X, s)
}

vector_prod_sum_sample <- function(X, Y, s) {
    .Call(`_dcortools_vector_prod_sum_sample`, X, Y, s)
}

dcovtermsmemvec <- function(X, Y, metrX, metrY, prmX = 0, prmY = 0, calcdcor = FALSE, calcperm = FALSE) {
    .Call(`_dcortools_dcovtermsmemvec`, X, Y, metrX, metrY, prmX, prmY, calcdcor, calcperm)
}

dcovtermsmem <- function(X, Y, metrX, metrY, prmX = 0, prmY = 0, calcdcor = FALSE, calcperm = FALSE) {
    .Call(`_dcortools_dcovtermsmem`, X, Y, metrX, metrY, prmX, prmY, calcdcor, calcperm)
}

dvartermsmemvec <- function(X, metrX, prmX = 0) {
    .Call(`_dcortools_dvartermsmemvec`, X, metrX, prmX)
}

dvartermsmem <- function(X, metrX, prmX = 0) {
    .Call(`_dcortools_dvartermsmem`, X, metrX, prmX)
}

aijbijmemvec <- function(X, Y, metrX, metrY, prmX = 0, prmY = 0) {
    .Call(`_dcortools_aijbijmemvec`, X, Y, metrX, metrY, prmX, prmY)
}

aijbijmem <- function(X, Y, metrX, metrY, prmX = 0, prmY = 0) {
    .Call(`_dcortools_aijbijmem`, X, Y, metrX, metrY, prmX, prmY)
}

SUMAIJBIJ <- function(IY, X, Y, XY, SX_X, SX_Y, SX_XY, SY_X, SY_Y, SY_XY) {
    .Call(`_dcortools_SUMAIJBIJ`, IY, X, Y, XY, SX_X, SX_Y, SX_XY, SY_X, SY_Y, SY_XY)
}

